@@ -7,12 +7,13 @@ from more_itertools import roundrobin
 from itertools import chain
 from autorepr import autorepr, autotext
 from multiprocessing import Manager
-from ast import parse, Call, Name, UnaryOp, Constant, List, Dict, Return, For, Assign, If, Load, Store, keyword, Compare, BinOp, BoolOp, Add, Sub, Div, Mult, FloorDiv, Mod, MatMult, BitAnd, BitOr, BitXor, Invert, Not, UAdd, USub, LShift, RShift, And, Or, Eq, NotEq, Lt, Gt, LtE, GtE, Eq, NotEq, In, NotIn, Is, IsNot, ImportFrom, Attribute, IfExp, Subscript, Index
+from ast import parse, Call, Name, UnaryOp, Constant, List, Dict, Return, For, Assign, If, Load, Store, keyword, Compare, BinOp, BoolOp, Add, Sub, Div, Mult, FloorDiv, Mod, MatMult, BitAnd, BitOr, BitXor, Invert, Not, UAdd, USub, LShift, RShift, And, Or, Eq, NotEq, Lt, Gt, LtE, GtE, Eq, NotEq, In, NotIn, Is, IsNot, ImportFrom, Attribute, IfExp, Subscript, Index, Tuple, Starred
+import ast
 import logging
 from importlib import import_module
 import builtins
 from tx.functional.either import Left, Right, Either
-from .dependentqueue import DependentQueue, SubQueue
+from .dependentqueue import DependentQueue
 from .stack import Stack
 from tx.readable_log import getLogger
 
@@ -28,7 +29,9 @@ def to_mod(value):
 
     
 def python_ast_to_value(expr):
-    if isinstance(expr, List):
+    if expr is None:
+        return None
+    elif isinstance(expr, List):
         return [python_ast_to_value(elt) for elt in expr.elts]
     elif isinstance(expr, Dict):
         return {python_ast_to_value(k): python_ast_to_value(v) for k, v in zip(expr.keys, expr.values)}
@@ -38,7 +41,7 @@ def python_ast_to_value(expr):
         raise RuntimeError(f"cannot convert ast {expr} to value")
 
     
-def python_ast_to_arg(expr):
+def python_ast_to_term(expr):
     if isinstance(expr, Name):
         return {
             "name": expr.id
@@ -61,7 +64,7 @@ def let_spec(body, func):
             spec_wrapped_by_assign = {
                 "type": "let",
                 "var": assign.targets[0].id,
-                "obj": python_ast_to_arg(assign.value),
+                "obj": python_ast_to_term(assign.value),
                 "sub": spec
             }
             return let_spec_handle_assign(resta, spec_wrapped_by_assign)
@@ -72,6 +75,16 @@ def extract_expressions_to_assignments(stmts, counter=[]):
     return list(chain(*(extract_expressions_to_assignments_in_statement(stmt, counter + [i]) for i, stmt in enumerate(stmts))))
 
 
+def extract_assignments_from_destructure(target, value, type_comments, counter):
+    if isinstance(target, Name):
+        return [Assign([target], value, type_comments)]
+    elif isinstance(target, Tuple):
+        name = generate_variable_name(counter)
+        return [Assign([Name(id=name, ctx=Store())], value)] + list(chain(*(extract_assignments_from_destructure(elt, Subscript(value=Name(id=name, ctx=Load()), slice=Index(value=Constant(value=i)), ctx=Load()), None, counter + [i]) for i, elt in enumerate(target.elts))))
+    else:
+        raise RuntimeError(f"unsupported assignment to {target}")
+    
+    
 # assuming that var with name _var[x] is not used
 # todo: check for var name
 def extract_expressions_to_assignments_in_statement(stmt, counter):
@@ -87,37 +100,26 @@ def extract_expressions_to_assignments_in_statement(stmt, counter):
         return assigns + [stmt_eta]
     elif isinstance(stmt, Return):
         ret = stmt.value
-        exprs, assigns = zip(*(extract_expressions_to_assignments_in_expression(value, counter) for value in ret.values))
-        stmt_eta = Return(value=Dict(keys=ret.keys, values=list(exprs)))
-        return list(chain(*assigns)) + [stmt_eta]
+        expr, assigns = extract_expressions_to_assignments_in_expression(stmt.value, counter)
+        stmt_eta = Return(value=expr)
+        return assigns + [stmt_eta]
     else:
         expr, assigns = extract_expressions_to_assignments_in_expression(stmt.value, counter, in_assignment=True)
-        stmt_eta = Assign(targets=stmt.targets, value=expr, type_comment=stmt.type_comment)
-        return assigns + [stmt_eta]
+        stmt_eta = extract_assignments_from_destructure(stmt.targets[0], expr, stmt.type_comment, counter + ["target"])
+        return assigns + stmt_eta
     
     
 def extract_expressions_to_assignments_in_expression(expr, counter, in_assignment=False):
     if isinstance(expr, Call):
-        if len(expr.args) == 0:
-            args = expr.args
-            assigns = []
-        else:
-            args, assign_lists = zip(*(extract_expressions_to_assignments_in_expression(arg, counter + [i]) for i, arg in enumerate(expr.args)))
-            assigns = list(chain(*assign_lists))
-        if len(expr.keywords) == 0:
-            keywords_exprs = expr.keywords
-            assigns_keywords = []
-        else:
-            keywords_exprs, assign_lists_keywords = zip(*(extract_expressions_to_assignments_in_expression(kw.value, counter + [kw.arg]) for kw in expr.keywords))
-            assigns_keywords = list(chain(*assign_lists_keywords))
-            
-#        logger.info(f"expr.args = {expr.args}\nargs = {args}\nassigns = {assigns}\nexpr.keywords = {expr.keywords}\nkeywords_exprs = {keywords_exprs}\nassigns_keywords = {assigns_keywords}")
-
+        args, assign_lists = extract_expressions_to_assignments_in_expressions(expr.args, counter)
+        assigns = list(chain(*assign_lists))
+        keywords_exprs, assign_lists_keywords = extract_expressions_to_assignments_in_expressions([(kw.arg, kw.value) for kw in expr.keywords], counter, keyword=True)
+        assigns_keywords = list(chain(*assign_lists_keywords))
         expr_eta = Call(func=expr.func, ctx=Load(), args=list(args), keywords=[keyword(arg=kw.arg, value=kw_expr) for kw, kw_expr in zip(expr.keywords, keywords_exprs)])
         assigns = assigns + assigns_keywords
     elif isinstance(expr, Compare):
         left, assigns_left = extract_expressions_to_assignments_in_expression(expr.left, counter + ["left"])
-        comparators, assign_lists = zip(*(extract_expressions_to_assignments_in_expression(comparator, counter + ["comparator", i]) for i, comparator in enumerate(expr.comparators)))
+        comparators, assign_lists = extract_expressions_to_assignments_in_expressions(expr.comparators, counter + ["comparator"])
         expr_eta = Compare(left=left, ops=expr.ops, comparators=list(comparators))
         assigns = assigns_left + list(chain(*assign_lists))
     elif isinstance(expr, BinOp):
@@ -126,7 +128,7 @@ def extract_expressions_to_assignments_in_expression(expr, counter, in_assignmen
         expr_eta = BinOp(left=left, op=expr.op, right=right)
         assigns = assigns_left + assigns_right
     elif isinstance(expr, BoolOp):
-        values, assign_lists = zip(*(extract_expressions_to_assignments_in_expression(value, counter + [i]) for i, value in enumerate(expr.values)))
+        values, assign_lists = extract_expressions_to_assignments_in_expressions(expr.values, counter)
         assigns = list(chain(*assign_lists))
         expr_eta = BoolOp(op=expr.op, values=list(values))
     elif isinstance(expr, UnaryOp):
@@ -143,17 +145,66 @@ def extract_expressions_to_assignments_in_expression(expr, counter, in_assignmen
         slice, assigns_slice = extract_expressions_to_assignments_in_expression(expr.slice.value, counter + ["slice.value"])
         expr_eta = Subscript(value=value, slice=Index(value=slice), ctx=expr.ctx)
         assigns = assigns_value + assigns_slice
+    elif isinstance(expr, Starred):
+        value, assigns = extract_expressions_to_assignments_in_expression(expr.value, counter)
+        expr_eta = Starred(value=value, ctx=expr.ctx)
+    elif isinstance(expr, List):
+        exprs, assign_lists = extract_expressions_to_assignments_in_expressions(expr.elts, counter)
+        assigns = list(chain(*assign_lists))
+        if not is_dynamic_args(exprs) and len(assigns) == 0:
+            return expr, []
+        else:
+            expr_eta = List(elts=list(exprs), ctx=expr.ctx)
+    elif isinstance(expr, Tuple):
+        exprs, assign_lists = extract_expressions_to_assignments_in_expressions(expr.elts, counter)
+        assigns = list(chain(*assign_lists))
+        if not is_dynamic_args(exprs) and len(assigns) == 0:
+            return expr, []
+        else:
+            expr_eta = Tuple(elts=list(exprs), ctx=expr.ctx)
+    elif isinstance(expr, Dict):
+        exprs_keys, assign_keys = extract_expressions_to_assignments_in_expressions(expr.keys, counter + ["keys"])
+        exprs_values, assign_values = extract_expressions_to_assignments_in_expressions(expr.values, counter + ["values"])
+        assigns = list(chain(*assign_keys, *assign_values))
+        if not is_dynamic_args(exprs_keys) and not is_dynamic_args(exprs_values) and len(assigns) == 0:
+            return expr, []
+        else:
+            expr_eta = Dict(keys=list(exprs_keys), values=list(exprs_values))
     else:
         return expr, []
+
+    logger.info(f"{ast.dump(expr_eta)} {[ast.dump(assign) for assign in assigns]}")
     if in_assignment:
-#       logger.info("in assignment")
+#       logger.debug("in assignment")
         return expr_eta, assigns
     else:
-#       logger.info("out of assignment")
-        a = generate_variable_name(counter)
-        return Name(id=a, ctx=Load()), assigns + [Assign(targets=[Name(id=a, ctx=Store())], value=expr_eta, type_comment=None)]
+#       logger.debug("out of assignment")
+        expr, assign = generate_expression_and_assignment(expr_eta, counter)
+        return expr, assigns + [assign]
             
 
+def extract_expressions_to_assignments_in_expressions(exprs, counter, in_assignment=False, keyword=False):
+    if len(exprs) == 0:
+            return exprs, []
+    else:
+        if keyword:
+            return zip(*(extract_expressions_to_assignments_in_expression(expr, counter + [i], in_assignment=in_assignment) for i, expr in exprs))
+        else:
+            return zip(*(extract_expressions_to_assignments_in_expression(expr, counter + [i], in_assignment=in_assignment) for i, expr in enumerate(exprs)))
+
+        
+def is_dynamic_args(exprs):
+    return any(is_dynamic(expr) for expr in exprs)
+
+
+def is_dynamic(expr):
+    return isinstance(expr, Name) or isinstance(expr, Call) or isinstance(expr, BinOp) or isinstance(expr, BoolOp) or isinstance(expr, UnaryOp) or isinstance(expr, Compare) or isinstance(expr, IfExp) or isinstance(expr, Subscript) or (isinstance(expr, List) and is_dynamic_args(expr.elts)) or (isinstance(expr, Tuple) and is_dynamic_args(expr.elts)) or (isinstance(expr, Dict) and (is_dynamic_args(expr.keys) or is_dynamic_args(expr.values))) or (isinstance(expr, Starred) and is_dynamic(expr.value))
+
+
+def generate_expression_and_assignment(expr_eta, counter):
+    a = generate_variable_name(counter)
+    return Name(id=a, ctx=Load()), Assign(targets=[Name(id=a, ctx=Store())], value=expr_eta, type_comment=None)
+
 def generate_variable_name(counter):
     return f"_var_{'_'.join(map(str, counter))}"
 
@@ -164,31 +215,26 @@ def python_to_spec(py):
 
     body_eta = extract_expressions_to_assignments(body)
     
-    return python_to_spec_seq(body_eta, EnvStack2())
+    return python_to_spec_seq(body_eta)
     
 
-def is_dynamic(stmt):
-    return isinstance(stmt, Call) or isinstance(stmt, BinOp) or isinstance(stmt, BoolOp) or isinstance(stmt, UnaryOp) or isinstance(stmt, Compare) or isinstance(stmt, IfExp) or isinstance(stmt, Subscript)
-
-
-def python_to_spec_seq(body, dep_set, imported_names = {}):
+def python_to_spec_seq(body, imported_names = {}):
     def func(stmts):
         importfroms = [stmt for stmt in stmts if isinstance(stmt, ImportFrom)]
         imported_names2 = {**imported_names, **{func : "" for func in dir(builtins)}, **{func : modname for importfrom in importfroms if (modname := importfrom.module) if (mod := import_module(modname)) if (names := importfrom.names) for func in (dir(mod) if any(x.name == "*" for x in names) else [alias.name for alias in names])}}
-        logger.info(f"imported_names2 = {imported_names2}")
+        logger.debug(f"imported_names2 = {imported_names2}")
         apps = [stmt for stmt in stmts if isinstance(stmt, Assign) and is_dynamic(stmt.value)]
         fors = [stmt for stmt in stmts if isinstance(stmt, For)]
         ifs = [stmt for stmt in stmts if isinstance(stmt, If)]
         returns = [stmt for stmt in stmts if isinstance(stmt, Return)]
-        dep_set2 = dep_set | {app.targets[0].id for app in apps}
     
-        return python_to_top_spec(apps + fors + ifs + returns, dep_set2, imported_names2)
+        return python_to_top_spec(apps + fors + ifs + returns, imported_names2)
     
     return let_spec(body, func)
 
 
-def python_to_top_spec(body, dep_set, imported_names):
-    specs = list(chain.from_iterable(python_to_spec_in_top(stmt, dep_set, imported_names) for stmt in body))
+def python_to_top_spec(body, imported_names):
+    specs = list(chain.from_iterable(python_to_spec_in_top(stmt, imported_names) for stmt in body))
     if len(specs) == 1:
         return specs[0]
     else:
@@ -197,50 +243,49 @@ def python_to_top_spec(body, dep_set, imported_names):
             "sub": specs
         }
 
-
 EnvStack2 = Stack(set())
 
-def python_ast_to_term(iter):
-    if isinstance(iter, Name):
-        coll_name = {
-            "name": iter.id
-        }
-    else:
-        coll_name = {
-            "data": python_ast_to_value(iter)
-        }
-    return coll_name
-
-
-def python_to_spec_in_top(stmt, dep_set, imported_names):
+def python_to_spec_in_top(stmt, imported_names):
     if isinstance(stmt, For):
         coll_name = python_ast_to_term(stmt.iter)
         return [{
             "type": "map",
             "var": stmt.target.id,
             "coll": coll_name,
-            "sub": python_to_spec_seq(stmt.body, EnvStack2(dep_set), imported_names)
+            "sub": python_to_spec_seq(stmt.body, imported_names)
         }]
     elif isinstance(stmt, If):
         cond_name = python_ast_to_term(stmt.test)
         return [{
             "type": "cond",
             "on": cond_name,
-            "then": python_to_spec_seq(stmt.body, EnvStack2(dep_set), imported_names),
-            "else": python_to_spec_seq(stmt.orelse, EnvStack2(dep_set), imported_names),
+            "then": python_to_spec_seq(stmt.body, imported_names),
+            "else": python_to_spec_seq(stmt.orelse, imported_names),
         }]
     elif isinstance(stmt, Return):
-        ret = stmt.value
+        ret_val = stmt.value
         return [{
             "type": "ret",
-            "var": ret_key.value,
             "obj": python_ast_to_term(ret_val)
-        } for ret_key, ret_val in zip(ret.keys, ret.values)]
+        }]
             
     else:
         targets = stmt.targets
-        name = targets[0].id
+        target = targets[0]
+        name = target.id
         app = stmt.value
+        if isinstance(app, Name):
+            func = "identity"
+            mod = "tx.functional.utils"
+            keywords = {
+                0: app
+            }
+        if isinstance(app, Starred):
+            func = "_starred"
+            mod = "tx.parallex.data"
+            keywords = {
+                0: app.value
+            }
         if isinstance(app, Call):
             fqfunc = app.func
             keywords = {
@@ -253,7 +298,7 @@ def python_to_spec_in_top(stmt, dep_set, imported_names):
             else:
                 func = fqfunc.id
                 mod = imported_names[func]
-            # logger.info(f"dep_set = {dep_set}")
+            # logger.debug(f"dep_set = {dep_set}")
         elif isinstance(app, Compare):
             if len(app.ops) > 1:
                 raise RuntimeError("unsupported multi-op compare")
@@ -359,7 +404,22 @@ def python_to_spec_in_top(stmt, dep_set, imported_names):
             mod = "tx.parallex.data"
             func = "_subscript"
                 
-        params = {k: python_ast_to_arg(v) for k, v in keywords.items()}
+        elif isinstance(app, List):
+            keywords = {i: elt for i, elt in enumerate(app.elts)}
+            mod = "tx.parallex.data"
+            func = "_list"
+                
+        elif isinstance(app, Tuple):
+            keywords = {i: elt for i, elt in enumerate(app.elts)}
+            mod = "tx.parallex.data"
+            func = "_tuple"
+                
+        elif isinstance(app, Dict):
+            keywords = {i: elt for i, elt in enumerate(app.keys + app.values)}
+            mod = "tx.parallex.data"
+            func = "_dict"
+                
+        params = {k: python_ast_to_term(v) for k, v in keywords.items()}
         
         return [{
             "type": "python",
