@@ -7,7 +7,6 @@ from importlib import import_module
 from itertools import chain
 from more_itertools import roundrobin
 from autorepr import autorepr, autotext
-from multiprocessing import Manager, Value
 from ast import parse, Call, Name, UnaryOp, Constant, List, Dict, Return, For
 import logging
 import traceback
@@ -18,10 +17,11 @@ from ctypes import c_int
 import builtins
 from tx.functional.either import Left, Right, Either
 from tx.functional.maybe import Just, Nothing, maybe
-from .dependentqueue import DependentQueue, SubQueue
+from .dependentqueue import DependentQueue
 from .utils import inverse_function
 from .python import python_to_spec, EnvStack2
 from .stack import Stack
+import jsonpickle
 from tx.readable_log import format_message, getLogger
 
 logger = getLogger(__name__, logging.INFO)
@@ -42,7 +42,7 @@ class AbsTask:
 
 class BaseTask(AbsTask):
     def run(self, results, subnode_results, queue):
-        logger.debug(f"BaseTask.run: restuls = {results}")
+        logger.debug("BaseTask.run: restuls = %s", results)
         return mbind(self.baseRun, results, subnode_results, queue)
     
 # task_counter = Value(c_int, 0)
@@ -105,7 +105,7 @@ class DynamicMap(BaseTask):
 
     def baseRun(self, results, subnode_results, queue):
         hold_id = queue.put(Hold(), is_hold=True)
-        logger.debug(f"DynamicMap.baseRun: put hold task on queue {hold_id}")
+        logger.debug("DynamicMap.baseRun: put hold task on queue %s", hold_id)
         enqueue({
             "type": "map",
             "var": self.var,
@@ -115,7 +115,7 @@ class DynamicMap(BaseTask):
             "sub": self.spec
         }, {**self.data, **{name: subnode_results[node_id] for name, node_id in self.subnode_top.items()}}, queue, top=EnvStack(), ret_prefix=self.ret_prefix, execute_unreachable=True, hold={hold_id})
         queue.complete(hold_id, {}, Nothing)
-        logger.debug(f"DynamicMap.baseRun: remove hold task from queue {hold_id}")
+        logger.debug("DynamicMap.baseRun: remove hold task from queue %s", hold_id)
         return {}, None
 
         
@@ -133,9 +133,9 @@ class DynamicGuard(BaseTask):
     __str__, __unicode__ = autotext("DynamicCond({self.task_id} {self.cond_spec} {self.then_spec} {self.else_spec} {self.data} {self.subnode_top} {self.ret_prefix})")
 
     def baseRun(self, results, subnode_results, queue):
-        logger.debug(f"DynamicCond.baseRun: before hold")
+        logger.debug("DynamicCond.baseRun: before hold")
         hold_id = queue.put(Hold(), is_hold=True)
-        logger.debug(f"DynamicCond.baseRun: put hold task on queue {hold_id}")
+        logger.debug("DynamicCond.baseRun: put hold task on queue %s", hold_id)
         enqueue({
             "type": "cond",
             "on": {
@@ -145,67 +145,63 @@ class DynamicGuard(BaseTask):
             "else": self.else_spec
         }, {**self.data, **{name: subnode_results[node_id] for name, node_id in self.subnode_top.items()}}, queue, top=EnvStack(), ret_prefix=self.ret_prefix, execute_unreachable=True, hold={hold_id})
         queue.complete(hold_id, {}, Nothing)
-        logger.debug(f"DynamicCond.baseRun: remove hold task from queue {hold_id}")
+        logger.debug("DynamicCond.baseRun: remove hold task from queue %s", hold_id)
         return {}, None
     
         
 class DynamicRet(AbsTask):
-    def __init__(self, var, obj_spec, ret_prefix, task_id=None):
+    def __init__(self, obj_spec, ret_prefix, task_id=None):
         super().__init__(task_id=task_id)
-        self.var = var
         self.obj_spec = obj_spec
         self.ret_prefix = ret_prefix
 
-    __repr__ = autorepr(["task_id", "var", "obj_spec", "ret_prefix"])
-    __str__, __unicode__ = autotext("DynamicRet(task_id={self.task_id} var={self.var} obj_spec={self.obj_spec} ret_prefix={self.ret_prefix})")
+    __repr__ = autorepr(["task_id", "obj_spec", "ret_prefix"])
+    __str__, __unicode__ = autotext("DynamicRet(task_id={self.task_id} obj_spec={self.obj_spec} ret_prefix={self.ret_prefix})")
 
     def run(self, results, subnode_results, queue):
-        return {".".join(chain(map(str, self.ret_prefix), [self.var])): results[self.obj_spec]}, None
+        return {".".join(map(str, self.ret_prefix)): results[self.obj_spec]}, None
 
         
 class Ret(AbsTask):
-    def __init__(self, var, obj, ret_prefix, task_id=None):
+    def __init__(self, obj, ret_prefix, task_id=None):
         super().__init__(task_id=task_id)
-        self.var = var
         self.obj = obj
         self.ret_prefix = ret_prefix
 
-    __repr__ = autorepr(["task_id", "var", "obj", "ret_prefix"])
-    __str__, __unicode__ = autotext("Ret({self.task_id} {self.var} {self.obj} {self.ret_prefix})")
+    __repr__ = autorepr(["task_id", "obj", "ret_prefix"])
+    __str__, __unicode__ = autotext("Ret({self.task_id} {self.obj} {self.ret_prefix})")
 
     def run(self, results, subnode_results, queue):
-        return {".".join(map(str, self.ret_prefix + [self.var])): Right(self.obj)}, None
+        return {".".join(map(str, self.ret_prefix)): Right(self.obj)}, None
 
         
 class EndOfQueue(AbsTask):
     def __init__(self):
         super().__init__()
 
-        
-def dispatch(job_queue, worker_queues):
-    n = len(worker_queues)
-    while True:
-        jri = job_queue.get()
-        job, results, subnode_results, jid = jri
-        logger.debug(f"dispatching type(job) = {type(job)} results = {results} subnode_results = {subnode_results} jid = {jid}")
-        if isinstance(job, EndOfQueue):
-            for p in worker_queues:
-                p.put_in_subqueue(jri)
-            break
-        else:
-            base = choice(range(n))
-            done = False
-            for off in range(n):
-                i = base + off
-                if not worker_queues[i].full():
-                    worker_queues[i].put_in_subqueue(jri)
-                    done = True
-                    break
-            if not done:
-                worker_queues[base].put_in_subqueue(jri)
+    def __eq__(self, other):
+        return isinstance(other, EndOfQueue)
 
-    
-def mbind(job_run, params, subnode_results, sub_queue):
+        
+def write_to_disk(dqueue, path):
+    with open(path, "w") as db:
+        while True:
+            output = dqueue.get_next_output()
+            if output == Nothing:
+                break
+            else:
+                db.write(jsonpickle.encode(output.value) + "\n")
+
+
+def read_from_disk(path):
+    obj = {}
+    with open(path) as db:
+        for line in db:
+            obj.update(jsonpickle.decode(line))
+    return obj
+
+        
+def mbind(job_run, params, subnode_results, queue):
     resultv = {}
     subnode_resultv = {}
     for k, v in params.items():
@@ -225,16 +221,17 @@ def mbind(job_run, params, subnode_results, sub_queue):
                 subnode_resultv[k] = v.value
         else:
             # logger.debug(f"mbind: running {job_run}")
-            ret, resultj = job_run(resultv, subnode_resultv, sub_queue)
+            ret, resultj = job_run(resultv, subnode_resultv, queue)
             if not isinstance(resultj, Either):
                 resultj = Right(resultj)
     return ret, resultj
 
                         
-def work_on(sub_queue, library_paths):
+def work_on(queue, library_paths):
+    logger.debug("library_paths = %s", library_paths)
     sys.path.extend(library_paths)
     while True:
-        jri = sub_queue.get()
+        jri = queue.get()
         job, results, subnode_results, jid = jri
         if isinstance(job, EndOfQueue):
             break
@@ -244,10 +241,10 @@ def work_on(sub_queue, library_paths):
             #     "jid": jid,
             #     "params": results
             # }))
-            logger.debug(f"task begin {type(job)} {jid}")
+            logger.debug("task begin %s %s", type(job), jid)
             try:
-                logger.info(f"task begin {jid}")
-                ret, resultj = job.run(results, subnode_results, sub_queue)
+                logger.info("task begin %s", jid)
+                ret, resultj = job.run(results, subnode_results, queue)
             except Exception as e:
                 resultj = Left((str(e), traceback.format_exc()))
                 ret = {}
@@ -258,9 +255,9 @@ def work_on(sub_queue, library_paths):
             #     "resultj": resultj,
             #     "params": results
             # }))
-            logger.debug(f"task finish {type(job)} {jid}")
-            logger.info(f"task finish {jid}")
-            sub_queue.complete(jid, ret, Just(resultj))
+            logger.debug("task finish %s %s", type(job), jid)
+            logger.info(f"task finish %s", jid)
+            queue.complete(jid, ret, Just(resultj))
     
 
 def depends_on(top, v):
@@ -307,8 +304,7 @@ def get_task_depends_on(top, sub):
 
     
 def get_task_non_dependency_params(top, spec):
-    logger.debug(f"top = {top}, spec = {spec}")
-    return {k: v for k, v in spec.get("params", {}).items() if depends_on(top, v) is Nothing}
+    return {k: v for k, v in spec.get("params", {}).items() if depends_on(top, v) == Nothing}
 
 
 no_op = {
@@ -404,7 +400,7 @@ def generate_dependency_graph(graph, node_map, dep_set, return_ids, sub, parent_
     
 
 def sort_tasks(dep_set, subs):
-    logger.debug(f"sort_tasks: before: {subs}, dep_set = {dep_set}")
+    # logger.debug(f"sort_tasks: before: {subs}, dep_set = {dep_set}")
     copy = list(subs)
     subs_sorted = []
     visited = set(dep_set)
@@ -431,7 +427,7 @@ def sort_tasks(dep_set, subs):
                 dep += f"depends_on = {get_task_depends_on(dep_set2, task)}\n"
             raise RuntimeError(f"unresolved dependencies or cycle in depedencies graph {dep}")
 
-    logger.debug(f"sort_tasks: after: {subs_sorted}")
+    # logger.debug(f"sort_tasks: after: {subs_sorted}")
     return subs_sorted
 
 
@@ -518,27 +514,26 @@ def generate_tasks(spec, data, top=EnvStack(), ret_prefix=[], hold=set()):
 #        logger.debug(f"add task: {task.task_id} depends_on {dependencies}")
         yield task, set(dependencies.values()) | hold, set()
     elif ty == "ret":
-        var = spec["var"]
         obj_name = spec["obj"]
         if "name" in obj_name and obj_name["name"] in top:
             obj_spec = top[obj_name["name"]]
-            task = DynamicRet(var, obj_spec, ret_prefix)
+            task = DynamicRet(obj_spec, ret_prefix)
             dep = {obj_spec}
             yield task, dep | hold, set()
         else:
             obj = arg_spec_to_arg(data, obj_name)
-            yield Ret(var, obj, ret_prefix), hold, set()
+            yield Ret(obj, ret_prefix), hold, set()
     else:
         raise RuntimeError(f'unsupported spec type {ty}')
 
 
 def enqueue(spec, data, job_queue, top=EnvStack(), ret_prefix=[], execute_unreachable=False, hold=set()):
-    logger.debug(f"enqueue: data = {data}")
+    # logger.debug(f"enqueue: data = {data}")
 
     for what in generate_tasks(spec if execute_unreachable else remove_unreachable_tasks(spec), data, top=top, ret_prefix=ret_prefix, hold=hold):
         job, dependencies, subnode_dependencies = what
         job_id = job.task_id
-        logger.debug(f"add task {type(job)} {job_id}\ndepends_on = {dependencies}\nsubnode_depends_on = {subnode_dependencies}")
+        logger.debug("add task %s %s\ndepends_on = %s\nsubnode_depends_on = %s", type(job), job_id, dependencies, subnode_dependencies)
         job_queue.put(job, job_id=job_id, depends_on=dependencies, subnode_depends_on=subnode_dependencies)
 
 
